@@ -1,5 +1,7 @@
+import inspect
+import typing
 from enum import Enum
-from typing import Callable, Awaitable, AsyncGenerator, Dict, List, Optional
+from typing import Callable, Awaitable, AsyncGenerator, Dict, List, Optional, Type, TypeVar
 
 from aiohttp import ClientTimeout
 
@@ -17,21 +19,30 @@ class API:
     def __init__(self, clazz, func, method: RestMethod, content_type: str, is_instance_method: bool,
                  is_class_method: bool,
                  is_constructor: bool,
+                 on_disconnect: Optional[Callable[[], None]] = None,
                  timeout: Optional[ClientTimeout] = None,
                  expire_on_exit: bool = False, uuid_param: Optional[str] = None):
-        annotations = func.__annotations__ if not is_class_method else func.__func__.__annotations__
+        annotations = func.__annotations__
         self._clazz = clazz
+        self._on_disconnect = on_disconnect
         self._is_class_method = is_class_method
         self._is_instance_method = is_instance_method
         self._is_static = not self._is_class_method and not self._is_instance_method
         self._expire_obj = expire_on_exit
         self._func = func
-        self._real_func = func if not is_class_method else func.__func__
+        self._real_func = func
         self._method = method
         self._timeout = timeout or ClientTimeout()
         if 'return' not in annotations:
             raise TypeError(f"No annotation for return type in {func}")
         self._arg_annotations = {name: typ for name, typ in annotations.items() if name != 'return'}
+        if not func.__name__.startswith('_'):
+            sig = inspect.signature(func)
+            for name in self._arg_annotations:
+                if sig.parameters[name].kind == inspect.Parameter.VAR_POSITIONAL:
+                    self._arg_annotations[name] = typing.Tuple[int, None]
+                elif sig.parameters[name].kind == inspect.Parameter.VAR_KEYWORD:
+                    self._arg_annotations[name] = typing.Dict[str, self._arg_annotations[name]]
         self._async_arg_annotations = {
             name: typ for name, typ in self._arg_annotations.items()
             if typ in (bytes, AsyncChunkIterator, AsyncLineIterator)
@@ -40,10 +51,10 @@ class API:
             raise TypeError("At most one parameter can be and async iterator in a web_api request")
         self._return_type = annotations['return']
         self._has_streamed_response = False
-        if str(self._return_type).startswith('typing.AsyncIterator'):
+        if str(self._return_type).startswith('typing.AsyncIterator') or self._return_type == typing.AsyncIterator:
             self._return_type = self._return_type.__args__[0]
             self._has_streamed_response = True
-        elif str(self._return_type).startswith('typing.AsyncGenerator'):
+        elif str(self._return_type).startswith('typing.AsyncGenerator') or self._return_type == typing.AsyncGenerator:
             self._return_type = self._return_type.__args__[1]
             self._has_streamed_response = True
         self._content_type = content_type if not self.has_streamed_response else 'text/streamed; charset=x-user-defined'
@@ -100,6 +111,10 @@ class API:
     def method(self):
         return self._method
 
+    @property
+    def on_disconnect(self):
+        return self._on_disconnect
+
     @property
     def is_instance_method(self) -> bool:
         return self._is_instance_method
@@ -113,12 +128,12 @@ class API:
         return self._arg_annotations
 
     @property
-    def async_arg_annotations(self) -> Dict[str, str]:
+    def async_arg_annotations(self) -> Dict[str, Type]:
         return self._async_arg_annotations
 
     @property
-    def synchronous_arg_annotations(self) -> List[str]:
-        return [a for a in self._arg_annotations if a not in self._async_arg_annotations]
+    def synchronous_arg_annotations(self) -> Dict[str, Type]:
+        return {a: v for a, v in self._arg_annotations.items() if a not in self._async_arg_annotations}
 
     @property
     def return_type(self):
@@ -135,6 +150,22 @@ class API:
     def __call__(self, *args, **kwargs):
         return self._func(*args, **kwargs)
 
+    def update_varargs(self, kwargs: Dict[str, typing.Any]) -> [Dict[str, typing.Any], typing.Tuple[typing.Any]]:
+        """
+        takes list of kwargs for the call and transforms as needed for any varargs  or varkwargs in call
+        """
+        sig = inspect.signature(self._func)
+        updated = kwargs
+        varargs = tuple()
+        for name, param in sig.parameters.items():
+            if param.kind == inspect.Parameter.VAR_KEYWORD:
+                del updated[name]
+                updated.update(**kwargs[name])
+            elif param.kind == inspect.Parameter.VAR_POSITIONAL:
+                varargs = kwargs[name]
+                del updated[name]
+        return updated, varargs
+
 
 class APIDoc:
     """
@@ -255,3 +286,10 @@ class APIDoc:
         else:
             # noinspection PyProtectedMember
             return typ.__name__ if hasattr(typ, '__name__') else typ._name if hasattr(typ, '_name') else str(typ)
+
+
+V = TypeVar('V')
+
+
+def raise_is_abstract(v: V = None) -> V:
+    raise NotImplemented(f"Call to abstract function")
