@@ -25,7 +25,7 @@ def normalize_to_json_compat(val: Any) -> Any:
         json_data = {}
         for key, value in val.items():
             json_data[to_str(key)] = normalize_to_json_compat(value)
-    elif type(val) in [list] or (getattr(type(val), '_name', None) in ('List', )):
+    elif type(val) in [list, set, tuple] or (getattr(type(val), '_name', None) in ('List', 'Set', 'Tuple')):
         json_data = []
         for value in val:
             json_data.append(normalize_to_json_compat(value))
@@ -43,6 +43,7 @@ def normalize_from_json(json_data, typ) -> Any:
                 return json_data
             elif arg in (str, int, float):
                 continue
+            # noinspection PyBroadException
             try:
                 return normalize_from_json(json_data, arg)
             except Exception:
@@ -60,7 +61,7 @@ def normalize_from_json(json_data, typ) -> Any:
     elif typ == str:
         return json_data
     elif typ in (int, float):
-        return json_data
+        return typ(json_data)
     elif typ == bool:
         return str(json_data).lower() == 'true'
     elif getattr(typ, '_name', None) in ('Dict', 'Mapping'):
@@ -70,14 +71,15 @@ def normalize_from_json(json_data, typ) -> Any:
             for k, v in json_data.items()
         }
     elif getattr(typ, '_name', None) in ('List', ):
-        elem_typ = typ.__args__[0]
-        return [normalize_from_json(value, elem_typ) for value in json_data]
+        return [normalize_from_json(value, typ.__args__[0]) for index, value in enumerate(json_data)]
     elif getattr(typ, '_name', None) in ('Set', ):
-        elem_typ = typ.__args__[0]
-        return {normalize_from_json(value, elem_typ) for value in json_data}
+        return {normalize_from_json(value, typ.__args__[0]) for index, value in enumerate(json_data)}
     elif getattr(typ, '_name', None) in ('Tuple', ):
-        elem_typ = typ.__args__[0]
-        return tuple(normalize_from_json(value, elem_typ) for value in json_data)
+        if typ.__args__[-1] == type(None):  # var args
+            return tuple(normalize_from_json(value, typ.__args__[0]) for index, value in enumerate(json_data))
+        else:
+            return tuple(normalize_from_json(value, typ.__args__[index]) for index, value in enumerate(json_data))
+
     elif hasattr(typ, '__dataclass_fields__'):
         return typ(**{
             name: normalize_from_json(json_data[name], field.type)
