@@ -15,7 +15,7 @@ import types
 import uuid as uuid_pkg
 
 import logging
-from aiohttp import ClientTimeout
+from aiohttp import ClientTimeout, ClientConnectionError
 from aiohttp import web
 from aiohttp.web import (
     Application,
@@ -23,6 +23,7 @@ from aiohttp.web import (
     Response,
     StreamResponse,
 )
+from asyncio import CancelledError
 from contextlib import suppress
 from pathlib import Path
 from ssl import SSLContext
@@ -38,6 +39,8 @@ from typing import (
     Type,
 )
 
+from asynciomultiplexer import asynciomultiplexer
+
 from . import HTTPException
 from .conversions import from_str, to_str, normalize_from_json
 from .decorators import (
@@ -72,8 +75,8 @@ class WebApplication:
     Main class for running a WebApplication server. See the documentation for *aiohttp* for information on the various
     parameters.  Additional parameters for this class are:
 
-    :static_path: root path to where static files will be kept, mapped to a route "/static", if provided
-    :js_bundle_name: the name of the javascript file, **without** exetension, where the client-side API
+    :param static_path: root path to where static files will be kept, mapped to a route "/static", if provided
+    :param js_bundle_name: the name of the javascript file, **without** extension, where the client-side API
        will be generated, if provided
     """
     _context: Dict[Awaitable, Request] = {}
@@ -82,6 +85,39 @@ class WebApplication:
     _instance_methods: List[API] = []
     _all_methods: List[API] = []
 
+    class MainThread:
+        """
+        This is used to "undo" the threading in aiohttp.  It seems silly to combine threading and asyncio
+        by handling each request in its own response.  It is somewhat understandable to attempt to
+        de-conflict state perhaps(?), but overall is just silly IMHO.  This thread puts all invocations
+        of routes into one thread handle a main asyncio loop.  User needs to beware of stateful-ness in that
+        one asyncio task can change the state while another is in the middle of being awaited (but threading
+        really doesn't solve this anyway).
+        """
+        MAX_SIMULTANEOUS_REQUESTS = 1000
+
+        request_q = asynciomultiplexer.AsyncAdaptorQueue(MAX_SIMULTANEOUS_REQUESTS)
+
+        @classmethod
+        async def start(cls, coro: Awaitable, resp_q: asynciomultiplexer.AsyncAdaptorQueue):
+            await cls.request_q.put((coro, resp_q))
+
+        @classmethod
+        async def main(cls, initializer: Optional[Callable[[], None]] = None):
+            if initializer:
+                initializer()
+
+            async def task(coro: Awaitable, resp_q: asynciomultiplexer.AsyncAdaptorQueue):
+                try:
+                    resp = await coro
+                except Exception as e:
+                    resp = e
+                await resp_q.put(resp)
+
+            while True:
+                request = await cls.request_q.get(polling_interval=0.01)
+                asyncio.create_task(task(*request))
+
     class ObjectRepo:
         DEFAULT_OBJECT_EXPIRATION: int = 60*60*2   # in seconds = 1 hour
 
@@ -120,6 +156,7 @@ class WebApplication:
                  client_max_size: int = MAX_CLIENTS,
                  using_async: bool = True,
                  debug: Any = ..., ) -> None:  # mypy doesn't support ellipsis
+        self._main_task: Optional[asyncio.Task] = None
         if static_path is not None and not Path(static_path).exists():
             raise ValueError(f"Provided static path, {static_path} does not exist")
         self._static_path = static_path
@@ -230,7 +267,7 @@ class WebApplication:
 
         :param route: route to register under
         :param async_handler: the raw handler for handling an incoming Request and returning a Response
-        :param api: the high-level deocrated web_api that will be invoked by the handler
+        :param api: the high-level decorated web_api that will be invoked by the handler
         :param module: module that is holding the @web_api
         """
         if route in cls.routes_get or route in cls.routes_post:
@@ -264,12 +301,60 @@ class WebApplication:
         cls.callables_post[route] = api
         cls.module_mapping_post.setdefault(module, {})[route] = api
 
+    # noinspection PyProtectedMember
+    @classmethod
+    def preprocess_module(cls, module_: str):
+        if module_ not in sys.modules:
+            mod = importlib.import_module(module_)
+        else:
+            mod = sys.modules.get(module_)
+        for class_name, clazz in inspect.getmembers(mod, predicate=inspect.isclass):
+            from .client import WebInterface
+            if issubclass(clazz, WebInterface):
+                ancestors = clazz.__mro__
+                for ancestor in ancestors:
+                    if ancestor == clazz:
+                        continue
+                    methods = inspect.getmembers(ancestor, predicate=inspect.ismethod) + \
+                        inspect.getmembers(ancestor, predicate=inspect.isfunction)
+                    for method_name, method in methods:
+                        immediate = getattr(clazz, method_name)
+                        if hasattr(immediate, '__func__'):
+                            immediate = immediate.__func__
+                        if hasattr(method, '_bantam_web_api') and not hasattr(immediate, '_bantam_web_api'):
+                            web_api(content_type=method._bantam_web_api.content_type,
+                                    method=method._bantam_web_api.method,
+                                    is_constructor=method._bantam_web_api.is_constructor,
+                                    expire_obj=method._bantam_web_api.expire_object,
+                                    timeout=method._bantam_web_api.timeout,
+                                    uuid_param=method._bantam_web_api.uuid_param)(immediate)
+                            immediate._bantam_web_api._clazz = clazz
+                            # handles case of _expire !-> expire
+                            immediate._bantam_web_api._name = method._bantam_web_api.name
+                        elif hasattr(method, '_bantam_web_api'):
+                            has_diff_web_api = (
+                                method._bantam_web_api.method != immediate._bantam_web_api.method
+                                or method._bantam_web_api.content_type != immediate._bantam_web_api.content_type
+                                or method._bantam_web_api.is_constructor != immediate._bantam_web_api.is_constructor
+                            )
+                            if has_diff_web_api:
+                                raise ValueError(
+                                    f"Mismatch in @web_api specifications in {class_name}.{method_name}"
+                                    f" and {ancestor.__name__}.{method_name}"
+                                )
+                            elif method._bantam_web_api.arg_annotations != immediate._bantam_web_api.arg_annotations:
+                                raise TypeError(
+                                    f"Mismatch in name and/or type specifications between {class_name}.{method_name}"
+                                    f" and {ancestor.__name__}.{method_name}"
+                                )
+
     # noinspection PyProtectedMember
     async def start(self,
                     modules: List[str],
                     host: Optional[str] = None,
                     port: Optional[int] = None,
                     path: Optional[str] = None,
+                    initializer: Optional[Callable[[], None]] = None,
                     shutdown_timeout: float = 60.0,
                     ssl_context: Optional[SSLContext] = None,
                     backlog: int = 128,
@@ -284,6 +369,8 @@ class WebApplication:
         :param host: optional host of app, defaults to '127.0.0.1'
         :param port: optional port to listen on (TCP)
         :param path: path, if using UNIX domain sockets to listen on (cannot specify both TCP and domain parameters)
+        :param initializer: optional function (no params, no return) to call on first bring-up, inside the
+            thread associated with the app's asyncio loop
         :param shutdown_timeout: force shutdown if a shutdown call fails to take hold after this many seconds
         :param ssl_context: for HTTPS server; if not provided, will default to HTTP connection
         :param backlog: number of unaccepted connections before system will refuse new connections
@@ -295,23 +382,21 @@ class WebApplication:
             on Windows.
         """
         # noinspection PyProtectedMember
+        self._main_task = asyncio.create_task(self.MainThread.main(initializer))
         from aiohttp.web import _run_app as web_run_app
         for module in modules:
-            # ensures web api's are loaded:
-            mod = importlib.import_module(module)
+            self.preprocess_module(module)
+            if module not in sys.modules:
+                mod = importlib.import_module(module)
+            else:
+                mod = sys.modules.get(module)
             for api in list(self.module_mapping_get.get(module, {}).values()):
-                if api.name in ['_expire']:
+                if api.name in ['_expire', '_create']:
                     continue
-                if api.is_class_method:
-                    class_name, method_name = api._real_func.__qualname__.split('.')
-                    api._func = getattr(getattr(mod, class_name), method_name)
                 self._process_module_classes(mod, api)
             for api in self.module_mapping_post.get(module, {}).values():
-                if api.name in ['_expire']:
+                if api.name in ['_expire', '_create']:
                     continue
-                if api.is_class_method:
-                    class_name, method_name = api._real_func.__qualname__.split('.')
-                    api._func = getattr(getattr(mod, class_name), method_name)
                 self._process_module_classes(mod, api)
         allowed_get_routes = {}
         for module in [m for m in modules if m in self.module_mapping_get]:
@@ -340,7 +425,7 @@ class WebApplication:
                 if self._using_async:
                     JavascriptGeneratorAsync.generate(
                         out=out, skip_html=False,
-                        allowed_routes =list(self.routes_get.keys()) + list(self.routes_post.keys())
+                        allowed_routes=list(self.routes_get.keys()) + list(self.routes_post.keys())
                     )
                 else:
                     JavascriptGenerator.generate(out=out, skip_html=False)
@@ -356,7 +441,10 @@ class WebApplication:
         """
         Shutdown this server
         """
+        if self._main_task is not None:
+            self._main_task.cancel()
         if self._started:
+
             await self._web_app.shutdown()
             self._started = False
 
@@ -383,15 +471,15 @@ class WebApplication:
 
             # noinspection PyDecorator
             @staticmethod
-            async def _create(*args, **kargs) -> str:
-                if kargs.get('__uuid') or (api.uuid_param is not None and api.uuid_param in kargs):
-                    self_id = kargs[api.uuid_param] if api.uuuid_param is not None else kargs['__uuid']
-                    del kargs['__uuid']
+            async def _create(*args, **kwargs) -> clazz_:
+                if kwargs.get('__uuid') or (create_api.uuid_param is not None and create_api.uuid_param in kwargs):
+                    self_id = kwargs[create_api.uuid_param] if create_api.uuuid_param is not None else kwargs['__uuid']
+                    del kwargs['__uuid']
                     if self_id in self.ObjectRepo.instances:
                         raise HTTPException(404, f"UUid {self_id} already in use. uuid's must be unique")
                 else:
                     self_id = str(uuid_pkg.uuid4())
-                instance = clazz_(*args, **kargs)
+                instance = clazz_(*args, **kwargs)
                 if hasattr(clazz_, '__aenter__'):
                     await instance.__aenter__()
                 self.ObjectRepo.instances[self_id] = instance
@@ -402,7 +490,6 @@ class WebApplication:
                 clazz_._create.__annotations__ = clazz_.__init__.__annotations__ if hasattr(clazz_, '__init__') else {}
             else:
                 clazz._create.__annotations__ = {}
-            clazz_._create.__annotations__['return'] = str
 
             clazz_._create.__doc__ = clazz_.__init__.__doc__ if hasattr(clazz_, '__init__') else f"""
                 Create an instance of {clazz_.__name__} on server, returning a unique string id for the instance.
@@ -412,10 +499,12 @@ class WebApplication:
                 :return: unique string id of instance created
                 """
             clazz_._create.__qualname__ = f"{clazz_.__name__}._create"
+            clazz_._create.__annotations__['return'] = clazz_
             _create.__func__.__annotations__ = clazz_._create.__annotations__
+            _create.__annotations__ = clazz_._create.__annotations__
             # noinspection PyProtectedMember
             route_name = f"/{clazz_.__name__}/_create"
-            api = API(
+            create_api = API(
                 content_type='text/plain',
                 is_instance_method=False,
                 is_class_method=False,
@@ -424,7 +513,7 @@ class WebApplication:
                 func=_create.__func__,
                 clazz=clazz_
             )
-            self.module_mapping_get.setdefault(clazz_.__module__)[route_name] = api
+            self.module_mapping_get.setdefault(clazz_.__module__, {})[route_name] = create_api
             self.routes_get[route_name] = clazz._create
             # self._all_apis[route_name] = api
             self._func_wrapper(clazz, clazz_._create, is_class_method=False, is_instance_method=False,
@@ -454,8 +543,8 @@ class WebApplication:
 
             # noinspection PyProtectedMember
             self._func_wrapper(clazz, clazz_._expire, is_class_method=False, is_instance_method=True,
-                              expire_on_exit=False,
-                              is_constructor=False, method=RestMethod.GET, content_type="text/plain")
+                               expire_on_exit=False,
+                               is_constructor=False, method=RestMethod.GET, content_type="text/plain")
 
         for clazz in [cl for _, cl in inspect.getmembers(mod) if inspect.isclass(cl)]:
             # noinspection PyProtectedMember
@@ -468,14 +557,14 @@ class WebApplication:
                 self._class_instance_methods.setdefault(clazz, []).append(api)
                 self._instance_methods_class_map[api] = clazz
                 if api.is_constructor:
-                    if api._func.__annotations__.get('return') != clazz.__name__:
+                    if api._func.__annotations__.get('return') not in (clazz, clazz.__name__):
                         raise TypeError("@web_api's declared as constructors must return that class's type")
 
                 if api.expire_object:
-                    async def wrapped(this, *args, **kargs):
+                    async def wrapped(this, *args, **kwargs):
                         try:
                             # noinspection PyProtectedMember
-                            return await api._func(this, *args, **kargs)
+                            return await api._func(this, *args, **kwargs)
                         finally:
                             if this in self.ObjectRepo.by_instance:
                                 uuid = self.ObjectRepo.by_instance[this]
@@ -485,7 +574,7 @@ class WebApplication:
                     setattr(clazz, api.name, wrapped)
             elif method_found and api in self._all_apis:
                 if api.is_constructor:
-                    if api._func.__annotations__.get('return') != clazz.__name__:
+                    if api._func.__annotations__.get('return') not in (clazz, clazz.__name__):
                         raise TypeError("@web_api's declared as constructors must return that class's type")
                 self._class_instance_methods.setdefault(clazz, [])  # create an empty list of instance methods at least
 
@@ -497,6 +586,7 @@ class WebApplication:
                       expire_on_exit: bool,
                       method: RestMethod,
                       content_type: str,
+                      on_disconnect: Optional[Callable[[], None]] = None,
                       timeout: Optional[ClientTimeout] = None,
                       uuid_param: Optional[str] = None,
                       preprocess: Optional[PreProcessor] = None,
@@ -509,59 +599,72 @@ class WebApplication:
         :return: function back, having processed as a web api and registered the route
         """
         api = API(clazz, func, method=method, content_type=content_type, is_instance_method=is_instance_method,
-                  is_class_method=is_class_method,
+                  is_class_method=is_class_method, on_disconnect=on_disconnect,
                   is_constructor=is_constructor, expire_on_exit=expire_on_exit, uuid_param=uuid_param, timeout=timeout)
         func._bantam_web_api = api
         if hasattr(func, '__func__'):
             func.__func__._bantam_web_api = api
+            func = func.__func__
         if is_instance_method:
-            if not inspect.iscoroutinefunction(func) and not inspect.isasyncgenfunction(func):
-                raise ValueError("the @web_api decorator can only be applied to classes with public "
-                                 f"methods that are coroutines (async); see {api.qualname}")
             cls._instance_methods.append(api)
-        elif not is_class_method:
-            if not inspect.iscoroutinefunction(func) and not inspect.isasyncgenfunction(func):
-                raise ValueError("the @web_api decorator can only be applied to methods that are coroutines (async)")
-        else:
-            if not inspect.iscoroutinefunction(func.__func__) and not inspect.isasyncgenfunction(func.__func__):
-                raise ValueError("the @web_api decorator can only be applied to methods that are coroutines (async)")
+        if not inspect.iscoroutinefunction(func) and not inspect.isasyncgenfunction(func):
+            raise ValueError("the @web_api decorator can only be applied to classes with public "
+                             f"methods that are coroutines (async); see {api.qualname}")
         cls._all_methods.append(api)
         name = api.qualname.replace('__init__', 'create').replace('._expire', '.expire')
         name_parts = name.split('.')[-2:]
         route = '/' + '/'.join(name_parts)
+        if '__isabstractmethod__' in func.__dict__:
+            return func
 
         async def invoke(app: WebApplication, request: Request):
-            nonlocal preprocess, postprocess
-            try:
-                preprocess = preprocess or app.preprocessor
+
+            async def invoke_proper():
+                nonlocal preprocess, postprocess
                 try:
-                    addl_args = (preprocess(request) or {}) if preprocess else {}
-                except Exception as e:
+                    preprocess = preprocess or app.preprocessor
+                    try:
+                        addl_args = (preprocess(request) or {}) if preprocess else {}
+                    except Exception as e:
+                        text = traceback.format_exc()
+                        resp_ = Response(status=400, text=f"Error in preprocessing request: {e}\n{text}")
+                        await resp_.prepare(request)
+                        return resp_
+                    if method == RestMethod.GET:
+                        # noinspection PyProtectedMember
+                        response = await cls._invoke_get_api_wrapper(
+                            api, content_type=content_type, request=request, **addl_args
+                        )
+                    elif method == RestMethod.POST:
+                        # noinspection PyProtectedMember
+                        response = await cls._invoke_post_api_wrapper(
+                            api, content_type=content_type, request=request, **addl_args
+                        )
+                    else:
+                        raise ValueError(f"Unknown method {method} in @web-api")
+                    try:
+                        postprocess = postprocess or app.postprocessor
+                        postprocess(response) if postprocess else response
+                    except Exception as e:
+                        text = traceback.format_exc()
+                        resp_ = Response(status=400, text=f"Error in post-processing of response: {e}\n{text}")
+                        await resp_.prepare(request)
+                        return resp_
+                    return response
+                except (CancelledError, ConnectionResetError, ClientConnectionError):
+                    raise
+                except BaseException as e:
                     text = traceback.format_exc()
-                    return Response(status=400, text=f"Error in preprocessing request: {e}\n{text}")
-                if method == RestMethod.GET:
-                    # noinspection PyProtectedMember
-                    response = await cls._invoke_get_api_wrapper(
-                        api, content_type=content_type, request=request, **addl_args
-                    )
-                elif method == RestMethod.POST:
-                    # noinspection PyProtectedMember
-                    response = await cls._invoke_post_api_wrapper(
-                        api, content_type=content_type, request=request, **addl_args
-                    )
-                else:
-                    raise ValueError(f"Unknown method {method} in @web-api")
-                try:
-                    postprocess = postprocess or app.postprocessor
-                    postprocess(response) if postprocess else response
-                except Exception as e:
-                    text = traceback.format_exc()
-                    return Response(status=400, text=f"Error in post-processing of response: {e}\n{text}")
-                return response
-            except BaseException as e:
-                text = traceback.format_exc()
-                return Response(status=500, reason=f"Server error: {e}",
-                                body=f"Server error: {e}\n{text}", content_type="test/plain")
+                    resp_ = Response(status=500, reason=f"Server error: {e}",
+                                     body=f"Server error: {e}\n{text}", content_type="test/plain")
+                    await resp_.prepare(request)
+                    return resp_
+            resp_q = asynciomultiplexer.AsyncAdaptorQueue(1)
+            await cls.MainThread.start(invoke_proper(), resp_q)
+            resp = await resp_q.get()
+            if isinstance(resp, Exception):
+                raise
+            return resp
 
         invoke.clazz = WebApplication._instance_methods_class_map.get(api) if is_instance_method else None
         if method == RestMethod.GET:
@@ -598,10 +701,12 @@ class WebApplication:
                     encoding = item.replace('charset=', '')
             # report first param that doesn't match the Python signature:
             for k in [p for p in request.query if p not in api.arg_annotations and p != 'self']:
-                return Response(
+                resp = Response(
                     status=400,
                     text=f"No such parameter or missing type hint for param {k} in method {api.qualname}"
                 )
+                await resp.prepare(request)
+                return resp
 
             # convert incoming str values to proper type:
             kwargs = {k: _convert_request_param(v, api.arg_annotations[k]) if k != 'self' else v
@@ -618,12 +723,21 @@ class WebApplication:
                 if instance is None:
                     raise ValueError(f"No instance found for request with 'self' id of {self_id}")
                 del kwargs['self']
-                result = api(instance, **kwargs)
+                kwargs, varargs = api.update_varargs(kwargs)
+                result = api(instance, *varargs, **kwargs)
             elif api.is_class_method:
-                result = api(api.__self__, *kwargs)
+                if isinstance(api.clazz, tuple):
+                    module_name, class_name = api.clazz
+                    api._clazz = getattr(sys.modules.get(module_name), class_name)
+                kwargs, varargs = api.update_varargs(kwargs)
+                if 'cls' not in kwargs:
+                    result = api(api.clazz, *varargs, **kwargs)
+                else:
+                    result = api(*varargs, **kwargs)
             else:
+                kwargs, varargs = api.update_varargs(kwargs)
                 # call the underlying function:
-                result = api(**kwargs)
+                result = api(*varargs, **kwargs)
             if inspect.isasyncgen(result):
                 #################
                 #  streamed response through async generator:
@@ -637,10 +751,21 @@ class WebApplication:
                     # iterate to get the one (and hopefully only) yielded element:
                     # noinspection PyTypeChecker
                     async for res in result:
-                        serialized = _serialize_return_value(res, encoding)
-                        if not isinstance(res, bytes):
-                            serialized += b'\n'
-                        await response.write(serialized)
+                        try:
+                            serialized = _serialize_return_value(res, encoding)
+                            if not isinstance(res, bytes):
+                                serialized += b'\0'
+                            await response.write(serialized)
+                        except (CancelledError, ConnectionResetError, ClientConnectionError):
+                            if api.on_disconnect is not None:
+                                if inspect.iscoroutinefunction(api.on_disconnect):
+                                    await api.on_disconnect(res)
+                                else:
+                                    api.on_disconnect(res)
+                            raise
+
+                except (CancelledError, ConnectionResetError, ClientConnectionError):
+                    raise
                 except Exception as e:
                     print(str(e))
                     raise asyncio.CancelledError(f"Error in server-side logic: {e}") from e
@@ -669,16 +794,33 @@ class WebApplication:
                         uuid, cls.ObjectRepo.DEFAULT_OBJECT_EXPIRATION))
                 else:
                     result = _serialize_return_value(result, encoding)
-                return Response(status=200, body=result if result is not None else b"Success",
+                resp = Response(status=200, body=result if result is not None else b"Success",
                                 content_type=content_type)
+                await resp.prepare(request)
+                return resp
+        except PermissionError as e:
+            resp = Response(status=401,
+                            reason=f"Improperly formatted query: {str(e)}\n{traceback.format_exc()}",
+                            text=f"Improperly formatted query: {str(e)}\n{traceback.format_exc()}",
+                            )
+            await resp.prepare(request)
+            return resp
         except TypeError as e:
-            return Response(status=400, text=f"Improperly formatted query: {str(e)}\n{traceback.format_exc()}")
-        except HTTPException as e:
-            return Response(status=e.status_code, text=f"{e}: \n{traceback.format_exc()}")
-        except (asyncio.CancelledError, asyncio.TimeoutError):
+            resp = Response(status=400,
+                            reason=f"Improperly formatted query: {str(e)}\n{traceback.format_exc()}",
+                            text=f"Improperly formatted query: {str(e)}\n{traceback.format_exc()}",
+                            )
+            await resp.prepare(request)
+            return resp
+        except HTTPException:
             raise
         except Exception as e:
-            return Response(status=500, text=f"Exception: {e}: \n{traceback.format_exc()}")
+            resp = Response(status=400,
+                            reason=f"Exception: {e}: \n{traceback.format_exc()}",
+                            text=f"Exception: {e}: \n{traceback.format_exc()}",
+                            )
+            await resp.prepare(request)
+            return resp
         finally:
             # noinspection PyUnresolvedReferences,PyProtectedMember
             del cls._context[sys._getframe(0)]
@@ -694,6 +836,7 @@ class WebApplication:
         :param request: request to be processed
         :return: http response object
         """
+
         async def line_by_line_response(req: Request):
             reader = req.content
             chunk = True
@@ -735,9 +878,6 @@ class WebApplication:
                     kwargs = {key: streamed_bytes_arg_value(request)}
                 for k, v in kwargs.items():
                     assert type(v) == str, f"key {k}={v} is of type {str(type(v))}\n{kwargs}"
-                #kwargs = {k: _convert_request_param(v, api.arg_annotations[k]) if k != 'self' else v
-                #          for k, v in kwargs.items()}
-                # noinspection PyTypeChecker
                 kwargs.update({k: _convert_request_param(v, api.synchronous_arg_annotations[k]) if k != 'self' else v
                                for k, v in request.query.items() if k in api.synchronous_arg_annotations})
             else:
@@ -745,10 +885,12 @@ class WebApplication:
                 bytes_response = await request.read()
                 json_dict = json.loads(bytes_response.decode('utf-8'))
                 for k in [p for p in json_dict if p not in api.arg_annotations and p != 'self']:
-                    return Response(
+                    resp = Response(
                         status=400,
                         text=f"No such parameter or missing type hint for param {k} in method {api.qualname}"
                     )
+                    await resp.prepare(request)
+                    return resp
 
                 # convert incoming str values to proper type:
                 kwargs.update(dict(json_dict))
@@ -759,6 +901,8 @@ class WebApplication:
                 if k != 'self':
                     value_type = api.arg_annotations[k]
                     kwargs[k] = normalize_from_json(v, value_type)
+                else:
+                    kwargs[k] = v
             if api.is_instance_method:
                 self_id = kwargs.get('self')
                 if self_id is None:
@@ -769,17 +913,24 @@ class WebApplication:
                         self_id = json.loads(self_id)['uuid']
                         instance = cls.ObjectRepo.instances.get(self_id)
                         if instance is None:
-                            raise Exception()
-                    except:
-                        raise ValueError(f"No instance id found for request {self_id}")
+                            raise ValueError(f"No instance id found for request {self_id}")
+                    except Exception:
+                        raise ValueError(f"Error in json representation of an instance: {self_id}")
                 del kwargs['self']
-                awaitable = api(instance, **kwargs)
+                kwargs, varargs = api.update_varargs(kwargs)
+                awaitable = api(instance, *varargs, **kwargs)
                 if api.expire_object:
                     del cls.ObjectRepo.instances[self_id]
             elif api.is_class_method:
-                awaitable = api(api.__self__, **kwargs)
+
+                if isinstance(api.clazz, tuple):
+                    module_name, class_name = api.clazz
+                    api._clazz = getattr(sys.modules.get(module_name), class_name)
+                kwargs, varargs = api.update_varargs(kwargs)
+                awaitable = api(api.clazz, *varargs, **kwargs)
             else:
-                awaitable = api(**kwargs)
+                kwargs, varargs = api.update_varargs(kwargs)
+                awaitable = api(*varags, **kwargs)
             if inspect.isasyncgen(awaitable):
                 #################
                 #  streamed response through async generator:
@@ -793,31 +944,81 @@ class WebApplication:
                 try:
                     # iterate to get the one (and hopefully only) yielded element:
                     # noinspection PyTypeChecker
+                    count = 0
                     async for res in awaitable:
-                        serialized = _serialize_return_value(res, encoding)
-                        if not isinstance(res, bytes):
-                            serialized += b'\n'
-                        await response.write(serialized)
+                        try:
+                            serialized = _serialize_return_value(res, encoding)
+                            if not isinstance(res, bytes):
+                                serialized += b'\0'
+                            await response.write(serialized)
+                            count += 1
+                        except (CancelledError, ConnectionResetError, ClientConnectionError):
+                            if api.on_disconnect is not None:
+                                # noinspection PyUnboundLocalVariable
+                                args = (instance, res, ) if api.is_instance_method else (res, )
+                                if inspect.iscoroutinefunction(api.on_disconnect):
+                                    await api.on_disconnect(*args)
+                                else:
+                                    api.on_disconnect(*args)
+                            raise
+
+                except (CancelledError, ConnectionResetError, ClientConnectionError):
+                    raise
                 except Exception as e:
                     print(str(e))
                     await async_q.put(None)
-                    raise RuntimeError(f"Execption in server-side logic: {e}") from e
+                    resp = Response(status=500, body=f"Exception in server-side logic: {e}",
+                                    content_type='text/plain')
+                    await resp.prepare(request)
+                    return resp
                 await response.write_eof()
+                return response
             else:
                 #################
                 #  regular response
                 #################
-                result = _serialize_return_value(await awaitable, encoding)
-                return Response(status=200, body=result if result is not None else b"Success",
+                if api.is_constructor:
+                    instance = await awaitable
+                    if api.clazz and hasattr(api.clazz, '__aenter__'):
+                        await instance.__aenter__()
+                    if hasattr(api.clazz, 'jsonrepr'):
+                        repr_ = api.clazz.jsonrepr(instance)
+                        uuid = repr_.get(api.uuid_param)
+                        content_type = 'application/json'
+                        result = json.dumps(repr_)
+                    else:
+                        uuid = kwargs.get(api.uuid_param, str(uuid_pkg.uuid4()))
+                        result = json.dumps({'uuid': uuid})
+                    cls.ObjectRepo.instances[uuid] = instance
+                    cls.ObjectRepo.by_instance[instance] = uuid
+                    cls.ObjectRepo.expiry[uuid] = asyncio.create_task(cls.ObjectRepo.expire_obj(
+                        uuid, cls.ObjectRepo.DEFAULT_OBJECT_EXPIRATION))
+                    resp = Response(status=200, body=result if result is not None else b"Success",
+                                    content_type=content_type)
+                    await resp.prepare(request)
+                    return resp
+                else:
+                    result = _serialize_return_value(await awaitable, encoding)
+                resp = Response(status=200, body=result if result is not None else b"Success",
                                 content_type=content_type)
+                await resp.prepare(request)
+                return resp
 
         except TypeError as e:
-            return Response(status=400, text=f"Improperly formatted query: {str(e)}")
+            resp = Response(status=400, text=f"Improperly formatted query: {str(e)}")
+            await resp.prepare(request)
+            return resp
         except HTTPException as e:
-            return Response(status=e.status_code, text=str(e))
+            resp = Response(status=e.status_code, text=str(e))
+            await resp.prepare(request)
+            return resp
+        except (CancelledError, ConnectionResetError, ClientConnectionError):
+            raise
         except Exception as e:
             text = f"Exception: {e}\n {traceback.format_exc()}"
-            return Response(status=500, text=text)
+            resp = Response(status=500, text=text)
+            await resp.prepare(request)
+            return resp
         finally:
             # noinspection PyUnresolvedReferences,PyProtectedMember
             del cls._context[sys._getframe(0)]
@@ -843,7 +1044,7 @@ def _serialize_return_value(value: Any, encoding: str) -> bytes:
     """
     Serialize a Python value into bytes to return through a Rest API.  If a basic type such as str, int or float, a
     simple str conversion is done, then converted to bytes.  If more complex, the conversion will invoke the
-    'serialize' method of the value, raising TypeError if such a method does not exist or does not have a bare
+    '__str__' method of the value, raising TypeError if such a method does not exist or does not have a bare
     (no-parameter) signature.
 
     :param value: value to convert
